import time

import cv2
import numpy as np
import datetime



def kernels_init():
    kernels_Num = 8
    kernels = ['_' for i in range(kernels_Num)]
    kernels[0] = np.array([[1, 2, 1], [0, 0, 0], [-1, -2, -1]], dtype=int)
    kernels[1] = np.array([[2, 1, 0], [1, 0, -1], [0, -1, -2]], dtype=int)
    kernels[2] = np.array([[1, 0, -1], [2, 0, -2], [1, 0, -1]], dtype=int)
    kernels[3] = np.array([[0, -1, -2], [1, 0, -1], [2, 1, 0]], dtype=int)
    kernels[4] = np.array([[-1, -2, -1], [0, 0, 0], [1, 2, 1]], dtype=int)
    kernels[5] = np.array([[-2, -1, 0], [-1, 0, 1], [0, 1, 2]], dtype=int)
    kernels[6] = np.array([[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]], dtype=int)
    kernels[7] = np.array([[0, 1, 2], [-1, 0, 1], [-2, -1, 0]], dtype=int)
    return kernels_Num,kernels



def img_filter(kernels_Num, kernels, img):
    img_RGB = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)  # 转成RGB 方便后面显示
# 灰度化处理图像
    grayImage = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    gradients = ['_' for i in range(kernels_Num)]
    for i in range(kernels_Num):
        gradients[i] = cv2.filter2D(grayImage, cv2.CV_16S, kernels[i])#卷积
    #     """
    #     显示梯度图像  代码
    #     """
    #     cv2.normalize(gradients[i],gradients[i],0,255,cv2.NORM_MINMAX)# 归一化到0-255
    #     abs = cv2.convertScaleAbs(gradients[i]) # 应该是化为整形
    #     cv2.imshow(str(i),abs)
    # cv2.imshow('222',img)
    # cv2.waitKey(0)

    return grayImage,gradients,img_RGB



def get_agg_amp(kernels_Num, grayImage, gradients):
    angle_list = [270, 315, 0, 45, 90, 135, 180, 225]
    amplitude = np.full(grayImage.shape,0)
    angle = np.full(grayImage.shape,-64)

    for r in range(grayImage.shape[0]):
        pAmp = amplitude[r]
        pAng = angle[r]

        pGrad = ['_' for i in range(kernels_Num)]
        for i in range(kernels_Num):
            pGrad[i] = gradients[i][r]
        for c in range(grayImage.shape[1]):
            for i in range(kernels_Num):
                if (pAmp[c] < pGrad[i][c]):# 取对应位置不同角度梯度的最大值
                    pAmp[c] = pGrad[i][c]
                    pAng[c] = angle_list[i] # 取对应位置梯度的最大值所用的方向
    return angle_list,amplitude,angle


def get_single_contour(grayImage, amplitude, angle):
    edge = np.full(grayImage.shape,0)
    edge.astype('uint8')
    thres = 100 #阈值  设置最小幅度值
    for r in range(1, grayImage.shape[0]-1):
        pAmp1 = amplitude[r-1]
        pAmp2 = amplitude[r]
        pAmp3 = amplitude[r+1]

        pAng = angle[r]
        pEdge = edge[r]
        for c in range(1, grayImage.shape[1]-1):
            if (pAmp2[c] < thres):
                continue
            if pAng[c] == 270:
                if pAmp2[c] > pAmp1[c] and pAmp2[c] >= pAmp3[c]:
                    pEdge[c] = 255
            elif pAng[c] == 90:
                if pAmp2[c] >= pAmp1[c] and pAmp2[c] > pAmp3[c]:
                    pEdge[c] = 255
            elif pAng[c] == 315:
                if pAmp2[c] > pAmp1[c - 1] and pAmp2[c] >= pAmp3[c + 1]:
                    pEdge[c] = 255
            elif pAng[c] == 135:
                if pAmp2[c] >= pAmp1[c - 1] and pAmp2[c] > pAmp3[c + 1]:
                    pEdge[c] = 255
            elif pAng[c] == 0:
                if pAmp2[c] > pAmp2[c - 1] and pAmp2[c] >= pAmp2[c + 1]:
                    pEdge[c] = 255
            elif pAng[c] == 180:
                if pAmp2[c] >= pAmp2[c - 1] and pAmp2[c] > pAmp2[c + 1]:
                    pEdge[c] = 255
            elif pAng[c] == 45:
                if pAmp2[c] >= pAmp1[c + 1] and pAmp2[c] > pAmp3[c - 1]:
                    pEdge[c] = 255
            elif pAng[c] == 225:
                if pAmp2[c] > pAmp1[c + 1] and pAmp2[c] >= pAmp3[c - 1]:
                    pEdge[c] = 255
    return edge


"""
亚像素处理
"""
def get_contour_DXL(kernels_Num, grayImage, angle_list, amplitude, angle, edge):
    root2 = np.sqrt(2.0)
    tri_list = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
    for i in range(kernels_Num):
     tri_list[0][i] = np.cos(angle_list[i] * np.pi / 180.0)
	# sin前面的负号非常关键, 因为图像的y方向和直角坐标系的y方向相反
     tri_list[1][i] = -np.sin(angle_list[i] * np.pi / 180.0)
    vPts = []
    for r in range(1, grayImage.shape[0]-1):
        pAmp1 = amplitude[r - 1]
        pAmp2 = amplitude[r]
        pAmp3 = amplitude[r + 1]

        pAng = angle[r]
        pEdge = edge[r]
        for c in range(1, grayImage.shape[1]-1):
            if (pEdge[c]):
                nAngTmp = 0
                dTmp = 0
                if pAng[c] == 270:
                    nAngTmp = 0
                    dTmp = (pAmp1[c] - pAmp3[c]) / (pAmp1[c] + pAmp3[c] - 2 * pAmp2[c]) * 0.5
                # print([c + dTmp * tri_list[0][nAngTmp],r + dTmp * tri_list[1][nAngTmp]])
                elif pAng[c] == 90:
                    nAngTmp = 4
                    dTmp = -(pAmp1[c] - pAmp3[c]) / (pAmp1[c] + pAmp3[c] - 2 * pAmp2[c]) * 0.5
                elif pAng[c] == 315:
                    nAngTmp = 1
                    dTmp = (pAmp1[c - 1] - pAmp3[c + 1]) / (pAmp1[c - 1] + pAmp3[c + 1] - 2 * pAmp2[c]) * root2 * 0.5
                elif pAng[c] == 135:
                    nAngTmp = 5
                    dTmp = -(pAmp1[c - 1] - pAmp3[c + 1]) / (pAmp1[c - 1] + pAmp3[c + 1] - 2 * pAmp2[c]) * root2 * 0.5
                elif pAng[c] == 0:
                    nAngTmp = 2
                    dTmp = (pAmp2[c - 1] - pAmp2[c + 1]) / (pAmp2[c - 1] + pAmp2[c + 1] - 2 * pAmp2[c]) * 0.5
                elif pAng[c] == 180:
                    nAngTmp = 6
                    dTmp = -(pAmp2[c - 1] - pAmp2[c + 1]) / (pAmp2[c - 1] + pAmp2[c + 1] - 2 * pAmp2[c]) * 0.5
                elif pAng[c] == 45:
                    nAngTmp = 3
                    dTmp = (pAmp3[c - 1] - pAmp1[c + 1]) / (pAmp1[c + 1] + pAmp3[c - 1] - 2 * pAmp2[c]) * root2 * 0.5
                elif pAng[c] == 225:
                    nAngTmp = 7
                    dTmp = -(pAmp3[c - 1] - pAmp1[c + 1]) / (pAmp1[c + 1] + pAmp3[c - 1] - 2 * pAmp2[c]) * root2 * 0.5

                x = c + dTmp * tri_list[0][nAngTmp]
                y = r + dTmp * tri_list[1][nAngTmp]
                vPts.append([x,y])
    return vPts


import copy

def ordering_contours(Points):
    order_points = []
    while len(Points)!=0:
        order_ps = np.empty(0).astype(np.float16)
        min_P = np.array([-1.0,-1.0])
        # print(len(Points))
        point0 = copy.deepcopy(Points[0])
        order_ps = np.append(order_ps,point0).reshape(-1,2)
        while abs(min_P[1]-point0[1]) + abs(min_P[0]-point0[0])!=0:
            sign = [True]*len(Points)
            min_ = 10
            pointx = order_ps[-1]
            for i in range(0, len(Points)):
                if abs(Points[i][1]-pointx[1])>2 or abs(Points[i][0]-pointx[0])>2:
                    continue
                if abs(Points[i][1]-pointx[1]) + abs(Points[i][0]-pointx[0])<min_:
                    if abs(Points[i][1]-point0[1]) + abs(Points[i][0]-point0[0])==0 and\
                    len(order_ps)<5:
                        continue
                    min_ = abs(Points[i][1]-pointx[1]) + abs(Points[i][0]-pointx[0])
                    min_P = copy.deepcopy(Points[i])
                    j = i
            if abs(min_P[1]-pointx[1]) + abs(min_P[0]-pointx[0])!=0:
                order_ps = np.append(order_ps,min_P).reshape(-1,2)
                sign[j] = False
                Points = Points[sign]
        order_ps = np.delete(order_ps, [-1,-2]).reshape(-1,2)
        # print(len(order_ps))
        order_points.append(order_ps)
    return order_points




"""
输出亚像素坐标
"""


if __name__ == '__main__':
    # print("len(vPts)", len(vPts))
    start = time.process_time()
    kernels_Num, kernels = kernels_init()
    img = cv2.imread("C:\\Users\\star\\Desktop\\DXLtext.png") 
    grayImage, gradients, img_RGB = img_filter(kernels_Num, kernels, img)

    angle_list, amplitude, angle = get_agg_amp(kernels_Num, grayImage, gradients)
    
    """
    显示幅值图像  代码
    """
    # cv2.normalize(amplitude,amplitude,0,255,cv2.NORM_MINMAX)
    # abs = cv2.convertScaleAbs(amplitude)
    # cv2.imshow('amplitude',abs)
    # cv2.imshow('222',img)
    # cv2.waitKey(0)

    """
    显示角度图像  代码
    """
    # cv2.normalize(angle,angle,0,255,cv2.NORM_MINMAX)
    # abs = cv2.convertScaleAbs(angle)
    # cv2.imshow('angle',abs)
    # cv2.imshow('222',img)
    # cv2.waitKey(0)

    edge = get_single_contour(grayImage, amplitude, angle)
    """
    显示单像素图像
    """
    # edge_show = cv2.convertScaleAbs(edge)
    # cv2.imshow('edge',edge_show)
    # cv2.imwrite('edge.png', edge_show)
    # cv2.waitKey(0)

    vPts = get_contour_DXL(kernels_Num, grayImage, angle_list, amplitude, angle, edge)
    order_points= ordering_contours(vPts)
    print(len(order_points))

    # for x,y in vPts:
    #     print(x,y)

    endtime = time.process_time()
    print (str(endtime - start) + 's')
